/****************************************************************************************/
/*											*/
/* This program is free software; you can redistribute it and/or modify it under the	*/
/* terms of the GNU General Public License as published by the Free Software		*/
/* Foundation; either version 2 of the License, or (at your option) any later version.	*/
/*											*/
/* This program is distributed in the hope that it will be useful, but WITHOUT ANY	*/
/* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A	*/
/* PARTICULAR PURPOSE. See the GNU General Public License for more details.		*/
/*											*/
/* You should have received a copy of the GNU General Public License along with this	*/
/* program; (See "COPYING"). If not, If not, see <http://www.gnu.org/licenses/>.        */
/*											*/
/*--------------------------------------------------------------------------------------*/
/*											*/
/*  Copyright   Joerg Anders, TU Chemnitz, Fakultaet fuer Informatik, GERMANY           */
/*		ja@informatik.tu-chemnitz.de						*/
/*											*/
/*											*/
/****************************************************************************************/

#include "idiotseditor.h"
#include <locale.h>
#include <cairo.h>
#include "localization.h"
#include <dynarray.h>

#define ID_ZOOM_STEP 1.6
#define ID_MIN_ZOOM 1.0
#define ID_MAX_ZOOM 1400.0
#define ID_DEFAULT_ZOOM 120.0

#define SEP_LINE_THICK 2.0
#define BAR_LINE_THICK 2.0
#define NOTE_LINE_THICK 1.0
#define SEP_LINE_DIST  16.0
#define SEP_LINE_BORDER (0.2 * SEP_LINE_DIST)
#define ID_HEIGHT (120 - 4 + 1) * SEP_LINE_DIST


const char *NedIdiotsEditor::guiDescription =
"<ui>"
"  <menubar name='IdMainMenu'>"
"    <menu action='IdFileMenu'>"
"      <menuitem action='IdNew'/>"
"      <menuitem action='IdOpen'/>"
"      <separator/>\n"
"      <menuitem action='IdExit'/>"
"    </menu>"
"    <menu action='IdViewMenu'>"
"      <menuitem action='Idzoom_in'/>"
"      <menuitem action='Idzoom_out'/>"
"    </menu>"
"  </menubar>"
"  <toolbar name='IdToolBar'>\n"
"    <toolitem action='Idzoom_in'/>\n"
"    <toolitem action='Idzoom_out'/>\n"
"  </toolbar>\n"
"  <accelerator name='<control>plus' action='Idzoom_in'/>\n"
"  <accelerator name='<control>minus' action='Idzoom_out'/>\n"
"</ui>";


const GtkActionEntry NedIdiotsEditor::file_entries[] = {
  { "IdFileMenu", NULL, _("_File") },
  { "IdViewMenu", NULL, _("_View") },
  { "IdNew", GTK_STOCK_NEW, _("New"), NULL, _("New File"), NULL /* G_CALLBACK(new_file) */ },
  { "IdOpen", GTK_STOCK_OPEN, _("Open  ..."), "<control>O", _("Open a file"), NULL /* G_CALLBACK(restore_score) */ },
  { "IdExit", GTK_STOCK_QUIT, _("E_xit"), "<control>Q", _("Exit the program"), G_CALLBACK(close_i_edit)},
  {"Idzoom_in", GTK_STOCK_ZOOM_IN, "in", "<ctrl>plus", _("Zoom in Ctrl + +"),  G_CALLBACK(zoom_in)},
  {"Idzoom_out", GTK_STOCK_ZOOM_OUT, "out", "<ctrl>minus", _("Zoom out Ctrl + -"), G_CALLBACK(zoom_out)},
};

const char *NedIdiotsEditor::pitch_tab[] = { "E", "F", "F#", "G", "G#", "A", "A#", "B", "C",
       "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B", "C",
       "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B", "C",
       "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B", "C",
       "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B", "C",
       "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B", "C",
       "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B", "C",
       "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B", "C",
       "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B", "C",
       "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B", "C"};


NedIdiotsEditor::NedIdiotsEditor(bool with_triplets, bool with_single_triplets, bool with_alto_clef, int voice_num,
			int tempo_change_density, int volume_change_density, bool force_piano, bool force_one_staff) :
m_zoom_value(ID_DEFAULT_ZOOM) , m_notes(NULL) {
	GtkWidget *menubar;
	GtkWidget *drawing_hbox;
	GdkColor bgcolor;
	GtkWidget *main_vbox;
	GtkWidget *table;
	GtkWidget *label_instr;
	GError *error;
    	double min, max;
	m_main_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title (GTK_WINDOW (m_main_window), _("Idiot's editor"));
	gtk_window_set_default_size (GTK_WINDOW (m_main_window), 600, 400);

	m_menu_action_group = gtk_action_group_new ("IdMenuActions");
	gtk_action_group_set_translation_domain(m_menu_action_group, NULL);
	gtk_action_group_add_actions (m_menu_action_group, file_entries, G_N_ELEMENTS (file_entries), (void *) this);

	m_ui_manager = gtk_ui_manager_new ();
	gtk_ui_manager_insert_action_group (m_ui_manager, m_menu_action_group, 0);

	error = NULL;
	if (!gtk_ui_manager_add_ui_from_string (m_ui_manager, guiDescription, -1, &error)) {
		g_message ("building menus failed: %s", error->message);
		g_error_free (error);
		NedResource::Abort("error");
	}

	menubar = gtk_ui_manager_get_widget (m_ui_manager, "/IdMainMenu");
	table = gtk_notebook_new();

    m_drawing_area = gtk_drawing_area_new ();
    gtk_widget_set_size_request(m_drawing_area, 1000, (int) ID_HEIGHT);
    m_scrolled_window = gtk_scrolled_window_new (NULL, NULL);
    m_hadjust = gtk_scrolled_window_get_hadjustment(GTK_SCROLLED_WINDOW(m_scrolled_window));
    m_vadjust = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(m_scrolled_window));
    gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(m_scrolled_window), m_drawing_area);

    m_piano_area = gtk_drawing_area_new();
    gtk_widget_set_size_request (m_piano_area, 40, 100);

    bgcolor.pixel = 0;
    bgcolor.red = bgcolor.green =  bgcolor.blue =  0xffff;
    gtk_widget_modify_bg(GTK_WIDGET(m_drawing_area), GTK_STATE_NORMAL, &bgcolor);


	 GTK_WIDGET_SET_FLAGS(m_drawing_area, GTK_CAN_FOCUS);
	 GTK_WIDGET_SET_FLAGS(m_main_window, GTK_CAN_FOCUS);
	gtk_widget_add_events(m_main_window, GDK_KEY_PRESS_MASK );
	gtk_widget_add_events(m_drawing_area, GDK_KEY_PRESS_MASK );
	/*
	gtk_widget_add_events(m_drawing_area, GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK | GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK |
			GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK);
			*/
    g_signal_connect (m_drawing_area, "expose-event",
		      G_CALLBACK (handle_expose), (void *) this);
    g_signal_connect (m_piano_area, "expose-event",
		      G_CALLBACK (draw_piano), (void *) this);
    g_signal_connect (m_hadjust, "value-changed", 
		      G_CALLBACK (handle_scroll), (void *) this);
    g_signal_connect (m_vadjust, "value-changed", 
		      G_CALLBACK (handle_scroll), (void *) this);
		      /*
	
    g_signal_connect (m_drawing_area, "button-press-event",
		      G_CALLBACK (handle_button_press), (void *) this);
    g_signal_connect (m_drawing_area, "button-release-event",
		      G_CALLBACK (handle_button_release), (void *) this);
    g_signal_connect (m_drawing_area, "motion-notify-event",
		      G_CALLBACK (handle_motion), (void *) this);
    g_signal_connect (m_drawing_area, "size-allocate",
		      G_CALLBACK (size_change_handler), (void *) this);
    g_signal_connect (m_drawing_area, "key-press-event",
		      G_CALLBACK (key_press_handler), (void *) this);
    g_signal_connect (m_drawing_area, "key-release-event",
		      G_CALLBACK (key_release_handler), (void *) this);
    g_signal_connect (m_drawing_area, "enter-notify-event",
		      G_CALLBACK (window_enter_handler), (void *) this);
    g_signal_connect (m_drawing_area, "leave-notify-event",
		      G_CALLBACK (window_leave_handler), (void *) this);
    g_signal_connect (m_drawing_area, "scroll-event",
		      G_CALLBACK (scroll_event_handler), (void *) this);
    g_signal_connect (m_main_window, "delete-event",
		      G_CALLBACK (handle_delete), (void *) this);
    g_signal_connect (m_page_selector, "value-changed",
		      G_CALLBACK (handle_page_request), (void *) this);
		      */

	drawing_hbox = gtk_hbox_new (FALSE, 0);

	gtk_box_pack_start (GTK_BOX (drawing_hbox), m_piano_area, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (drawing_hbox), m_scrolled_window, TRUE, TRUE, 0);
	main_vbox = gtk_vbox_new (FALSE, 0);
	
	gtk_box_pack_start (GTK_BOX (main_vbox), menubar, FALSE, FALSE, 0);

	label_instr = gtk_label_new(_("Instr. 1"));
	gtk_notebook_append_page (GTK_NOTEBOOK(table), drawing_hbox, label_instr);
	gtk_box_pack_start (GTK_BOX (main_vbox), table, TRUE, TRUE, 0);
    gtk_container_add (GTK_CONTAINER (m_main_window), main_vbox);

    gtk_widget_show_all (m_main_window);
    g_object_get (m_vadjust, "lower", &min, "upper", &max, NULL);
     gtk_adjustment_set_value(GTK_ADJUSTMENT(m_vadjust), min + (max - min) / 2.0);

}

NedIdiotsEditor::~NedIdiotsEditor() {}


void NedIdiotsEditor::close_i_edit(GtkWidget  *widget, void *data) {
	NedIdiotsEditor *id_editor = (NedIdiotsEditor*) data;

	gtk_widget_destroy(id_editor->m_main_window);
}

void NedIdiotsEditor::closeNote(int pitch, unsigned long long end_time, int midi_nr) {
	GList *lptr;

	for (lptr = g_list_last(m_notes); lptr; lptr = g_list_previous(lptr)) {
		if (((NedIdNote *) lptr->data)->m_pitch == pitch &&
			!((NedIdNote *) lptr->data)->m_valid && ((NedIdNote *) lptr->data)->m_midi_start < end_time) {
			((NedIdNote *) lptr->data)->m_midi_stop = end_time;
			((NedIdNote *) lptr->data)->m_valid = true;
			repaint();
			return;
		}
	}
	printf("cannot close note\n"); fflush(stdout);
}

void NedIdiotsEditor::addNote(int pitch, unsigned long long  start, int midi_nr) {
	m_notes = g_list_append(m_notes, new NedIdNote(pitch, start));
}

gboolean NedIdiotsEditor::handle_expose (GtkWidget *widget, GdkEventExpose *event, gpointer data) {
	NedIdiotsEditor *id_editor = (NedIdiotsEditor*) data;
	id_editor->draw() ;
	
	return FALSE;
}

void NedIdiotsEditor::repaint() {
	gdk_window_invalidate_rect (m_drawing_area->window, NULL, FALSE);
	gdk_window_invalidate_rect (m_piano_area->window, NULL, FALSE);
}

void NedIdiotsEditor::handle_scroll(GtkAdjustment *adjustment, gpointer data) {
	NedIdiotsEditor *id_editor = (NedIdiotsEditor*) data;
	id_editor->repaint();
}

void NedIdiotsEditor::draw () {
	int i;
	double x, y;
	int width, height;
	cairo_t *cr;
	double d;
	double paper_long = 1000.0;
	double current_dist;
	int current_color;
	NedIdNote *note;
	GList *lptr;

#define MIN_ZOOM_DIST 8.0


	x = gtk_adjustment_get_value(m_hadjust);
	y = gtk_adjustment_get_value(m_vadjust);
	width = m_scrolled_window->allocation.width;
	height = m_scrolled_window->allocation.height;

	cr = gdk_cairo_create (m_drawing_area->window);

#define BAR_COLOR 0.0, 0.0, 0.0
#define HALF_COLOR 0.0, 0.0, 1.0
#define QUARTER_COLOR  0.0, 1.0, 0.0
#define EIGHTS_COLOR 0.5, 0.0, 0.5
#define COLOR_16 0.0, 0.5, 0.5
#define COLOR32  0.5, 0.5, 0.0
#define COLOR64 0.9, 0.2, 0.6

	static double color_array[7][3] = {
		{BAR_COLOR},
		{HALF_COLOR},
		{QUARTER_COLOR},
		{EIGHTS_COLOR},
		{COLOR_16},
		{COLOR32},
		{COLOR64}};



	cairo_new_path(cr);
	cairo_set_source_rgb (cr, 1.0, 1.0, 1.0);
	cairo_rectangle (cr,  x,  y, width, height);
	cairo_fill(cr);
	cairo_set_line_width(cr, SEP_LINE_THICK);
	for (i = 4; i < 4 + 117; i++) {
		cairo_new_path(cr);
		if (strlen(pitch_tab[i-4]) > 1) {
			cairo_new_path(cr);
			cairo_set_source_rgb (cr, 0.8, 0.8, 0.8);
			cairo_rectangle (cr, x, (i - 4) * SEP_LINE_DIST, width, SEP_LINE_DIST);
			cairo_fill(cr);
		}
		cairo_new_path(cr);
		cairo_set_source_rgb (cr, 0.0, 0.0, 0.0);
		if (y < (i - 3) * SEP_LINE_DIST  && y + height > (i - 4) * SEP_LINE_DIST) {
			cairo_move_to(cr, x, (i - 4) * SEP_LINE_DIST);
			cairo_line_to(cr, x + width, (i - 4) * SEP_LINE_DIST);
		}
		cairo_stroke(cr);
	}
	current_dist = m_zoom_value / 128.0;
	current_color = 6;
	while (current_color >= 0)  {
		if (current_dist > MIN_ZOOM_DIST) {
			cairo_new_path(cr);
			cairo_set_line_width(cr, current_color == 0 ? BAR_LINE_THICK : NOTE_LINE_THICK);
			cairo_set_source_rgb (cr, color_array[current_color][0], color_array[current_color][1],color_array[current_color][2]);
			for (d = 0.0; d < paper_long * m_zoom_value ; d += current_dist) {
				if (d < x) continue;
				if (d > x + width) break;
	
				cairo_move_to(cr, d, y);
				cairo_line_to(cr, d, y + height);
		
			}
			cairo_stroke(cr);
		}
		current_color--;
		current_dist *= 2.0;
	}
	cairo_set_source_rgb (cr, 1.0, 0.0, 0.0);
	for (lptr = g_list_first(m_notes); lptr; lptr = g_list_next(lptr)) {
		note = (NedIdNote *) lptr->data;
		if (!note->m_valid) continue;
		cairo_new_path(cr);
		cairo_rectangle (cr, note->m_midi_start / NOTE_4 * m_zoom_value / 4.0, (note->m_pitch - 4) * SEP_LINE_DIST + SEP_LINE_BORDER, 
				 (note->m_midi_stop - note->m_midi_start) / NOTE_4 * m_zoom_value / 4.0, SEP_LINE_DIST - 2 * SEP_LINE_BORDER);
		cairo_fill(cr);
	}
	cairo_destroy (cr);
}

gboolean NedIdiotsEditor::draw_piano (GtkWidget *widget, GdkEventExpose *event, gpointer data) {
	NedIdiotsEditor *id_editor = (NedIdiotsEditor*) data;
	int i;
	cairo_t *cr;
	double y;
	int width, height;

	y = gtk_adjustment_get_value(id_editor->m_vadjust);
	width = id_editor->m_scrolled_window->allocation.width;
	height = id_editor->m_scrolled_window->allocation.height;

#define T_X_OFFS 2.0


	cr = gdk_cairo_create (id_editor->m_piano_area->window);
	cairo_set_source_rgb (cr, 1.0, 2.0, 1.0);
	cairo_rectangle (cr,  0,  0, widget->allocation.width, widget->allocation.height);
	cairo_fill(cr);
	cairo_select_font_face (cr, "SANS", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);
	cairo_set_font_size(cr, 8.0);
	cairo_set_source_rgb (cr, 0.0, 0.0, 0.0);
	for (i = 4; i < 4 + 117; i++) {
		if (strlen(pitch_tab[i-4]) > 1) {
			cairo_new_path(cr);
			cairo_set_source_rgb (cr, 0.8, 0.8, 0.8);
			cairo_rectangle (cr, 0, (i - 4) * SEP_LINE_DIST - y, 100, SEP_LINE_DIST);
			cairo_fill(cr);
		}
		cairo_new_path(cr);
		cairo_set_source_rgb (cr, 0.0, 0.0, 0.0);
		if (y < (i - 3) * SEP_LINE_DIST  && y + height > (i - 4) * SEP_LINE_DIST) {
			cairo_move_to(cr, T_X_OFFS, (i - 4) * SEP_LINE_DIST - y + SEP_LINE_DIST / 1.8);
			cairo_show_text(cr, pitch_tab[i-4]);
		}
		cairo_stroke(cr);
	}
	cairo_destroy (cr);
	return FALSE;
}

void NedIdiotsEditor::zoom_in(GtkWidget  *widget, void *data) {
	NedIdiotsEditor *id_editor = (NedIdiotsEditor*) data;
	id_editor->m_zoom_value *= ID_ZOOM_STEP;
	if (id_editor->m_zoom_value > ID_MAX_ZOOM) {
		id_editor->m_zoom_value = ID_MAX_ZOOM;
	}
        gtk_widget_set_size_request(id_editor->m_drawing_area, (int) (1000.0 *  id_editor->m_zoom_value / ID_DEFAULT_ZOOM), (int) ID_HEIGHT);
	id_editor->repaint();
}

void NedIdiotsEditor::zoom_out(GtkWidget  *widget, void *data) {
	NedIdiotsEditor *id_editor = (NedIdiotsEditor*) data;
	id_editor->m_zoom_value /= ID_ZOOM_STEP;
	if (id_editor->m_zoom_value < ID_MIN_ZOOM) {
		id_editor->m_zoom_value = ID_MIN_ZOOM;
	}
        gtk_widget_set_size_request(id_editor->m_drawing_area, (int) (1000.0 * id_editor->m_zoom_value / ID_DEFAULT_ZOOM), (int) ID_HEIGHT);
	id_editor->repaint();
}
